package SearchTest;

use TWiki::Contrib::DBCacheContrib::Map;
use TWiki::Contrib::DBCacheContrib::Search;
use base qw(Test::Unit::TestCase);

sub new {
  my $self = shift()->SUPER::new(@_);
  $self->{map} = undef;
  return $self;
}

sub set_up {
  my $this = shift;
  $this->{map} = new TWiki::Contrib::DBCacheContrib::Map("
     string = \"String\",
     number = 99,
     date = \"3-Jul-1960\"");
  # date's a sunday!
  my $mother = new TWiki::Contrib::DBCacheContrib::Map("who=Mother");
  $this->{map}->set("mother", $mother);
  my $gran = new TWiki::Contrib::DBCacheContrib::Map("who=GrandMother");
  $mother->set("mother", $gran);
}

sub tear_down {
  my $this = shift;
  $this->{map} = undef;
}

sub test_empty {
  my $this = shift;
  my $search = create Search("");
  $this->assert_not_null( $search );
  $this->assert_equals(1,$search->matches($this->{map}));
}

sub test_badparse {
  my $this = shift;
  eval {  create Search("WITHIN_DAYS"); };
  $this->assert_not_null($@,$@);
  eval {  create Search("x WITHIN_DAYS 30"); };
  $this->assert_not_null($@,$@);
  eval {  create Search("z WITHIN_DAYS < 3"); };
  $this->assert_not_null($@,$@);
}

sub check {
  my ( $this, $s, $r ) = @_;
  my $search = create Search($s);
  my $res = $search->matches($this->{map});
  $this->assert_equals($r,
		       $res,
		       $search->toString().
		       " received $res, expected $r");

}

sub test_stringops {
  my $this = shift;
  $this->check("string='String'",1);
  $this->check("string='String '",0);
  $this->check("string=~'String '", 0);
  $this->check("string='Str'", 0);
  $this->check("string=~'trin'", 1);
  $this->check("string=~' String'", 0);
  $this->check("string!='Str'", 1);
  $this->check("string!='String '", 1);
  $this->check("string!='String'", 0);
}

sub test_numops {
  my $this = shift;
  $this->check("number='99'", 1);
  $this->check("number='98'", 0);
  $this->check("number=~'99'", 1);
  $this->check("number=~'98'", 0);
  $this->check("number!='99'", 0);
  $this->check("number!='0'", 99);
  $this->check("number<'100'", 1);
  $this->check("number<'99'", 0);
  $this->check("number>'98'", 1);
  $this->check("number>'99'", 0);
  $this->check("number<='99'", 1);
  $this->check("number<='100'", 1);
  $this->check("number<='98'", 0);
  $this->check("number>='98'", 1);
  $this->check("number>='99'", 1);
  $this->check("number>='100'", 0);
}

sub test_dateops {
  my $this = shift;
  $this->check("\$CALC(\"\$TIME(\$T(date))\") = \$CALC(\"\$TIME(3 Jul 1960)\")", 1);
  $this->check("\$CALC(\"\$TIME(\$T(date))\") = \$CALC(\"\$TIME(3-Jul-1960)\")", 1);
  $this->check("\$CALC(\"\$TIME(\$T(date))\") = \$CALC(\"\$TIME(4-Jul-1960)\")", 0);
  $this->check("\$CALC(\"\$TIME(\$T(date))\") < \$CALC(\"\$TIME(4 Jul 1960)\")", 1);
  $this->check("\$CALC(\"\$TIME(\$T(date))\") < \$CALC(\"\$TIME(3-Jul-1960)\")", 0);
  $this->check("\$CALC(\"\$TIME(\$T(date))\") < \$CALC(\"\$TIME(2-Jul-1960)\")", 0);
  $this->check("\$CALC(\"\$TIME(\$T(date))\") > \$CALC(\"\$TIME(2 Jul 1960)\")", 1);
  $this->check("\$CALC(\"\$TIME(\$T(date))\") > \$CALC(\"\$TIME(3-Jul-1960)\")", 0);
  $this->check("\$CALC(\"\$TIME(\$T(date))\") > \$CALC(\"\$TIME(4-Jul-1960)\")", 0);
}

sub test_dateops2 {
  my $this = shift;
  $this->check("\$CALC(\"\$WORKINGDAYS(\$TIME(29 Jun 1960), \$TIME(\$T(date)))\")", 3);
  $this->check("\$CALC(\"\$WORKINGDAYS(\$TIME(30 Jun 1960), \$TIME(\$T(date)))\")", 2);
  $this->check("\$CALC(\"\$WORKINGDAYS(\$TIME(1 Jul 1960), \$TIME(\$T(date)))\")", 1);
  $this->check("\$CALC(\"\$WORKINGDAYS(\$TIME(2 Jul 1960), \$TIME(\$T(date)))\")", 0);
  $this->check("\$CALC(\"\$WORKINGDAYS(\$TIME(\$T(date)), \$TIME(3 Jul 1960))\")", 0);
  $this->check("\$CALC(\"\$WORKINGDAYS(\$TIME(\$T(date)), \$TIME(4 Jul 1960))\")", 0);
  $this->check("\$CALC(\"\$WORKINGDAYS(\$TIME(\$T(date)), \$TIME(5 Jul 1960))\")", 1);
}

sub test_not {
  my $this = shift;
  $this->check("!number='99'",0);
  $this->check("!number='98'", 1);
}

sub test_and {
  my $this = shift;
  $this->check("number='99' AND string='String'", 1);
  $this->check("number='98' AND string='String'", 0);
  $this->check("number='99' AND string='Sring'", 0);
  $this->check("number='99' AND string='String' AND \$CALC(\"\$TIME(\$T(date))\") = \$CALC(\"\$TIME(3 Jul 1960)\")", 1);
}

sub test_or {
  my $this = shift;
  $this->check("number='99' OR string='Spring'",1);
  $this->check("number='98' OR string='String'", 1);
  $this->check("number='98' OR string='Spring'", 0);
}

sub conjoin {
  my ( $this, $last, $A, $B, $a, $b, $c, $r ) = @_;

  my $ae = "number='" . ( $a ? "99" : "98" ) . "'";
  my $be = "string='" . ( $b ? "String" : "Spring" ) . "'";
  my $ce = "number='" . ( $c ? "99" : "98" ) . "'";
  my $expr;
  if ( $last ) {
    $expr = "$ae $A ( $be $B $ce )";
  } else {
    $expr = "( $ae $A $be ) $B $ce";
  }
  $this->check($expr,$r);
}

sub test_brackets {
  my $this = shift;
  for (my $a = 0; $a < 2; $a++) {
    for (my $b = 0; $b < 2; $b++) {
      for (my $c = 0; $c < 2; $c++) {
	$this->conjoin(1,"AND","OR", $a, $b, $c, $a && ($b || $c));
	$this->conjoin(1,"OR","AND", $a, $b, $c, $a || ($b && $c));
	$this->conjoin(0,"AND","OR",$a, $b, $c, ($a && $b) || $c);
	$this->conjoin(0,"OR","AND",$a, $b, $c, ($a || $b) && $c);
      }
    }
  }
}

sub test_other {
  my $this = shift;
  $this->check("mother.who='Mother'",1);
  $this->check("mother.who!='Mother'",0);
  $this->check("mother.mother.who='GrandMother'",1);
}

sub test_calc {
  my $this = shift;
  $this->check("\$CALC(\"\$TRANSLATE(boom,bm,cl)\")","cool");
  $this->check("\$CALC(\"\$TRANSLATE(\$T(string),Si,su)\")","strung");
}

sub test_calc2 {
  my $this = shift;
  $this->check("\$CALC(\"\$TRANSLATE(\$T(mother.who),M,b)\")","bother");
  $this->check("\$CALC(\"\$TRANSLATE(\$T(mother.mother.who),GM,bb)\")","brandbother");
}

sub test_calc3 {
  my $this = shift;
  $this->check("\$CALC(\"\$TIME(13 Sep 2001)\")", 1000357200);
}

1;
